@@ -6,7 +6,7 @@
  */
 
 /* This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was ! distributed with this
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #ifdef _MSC_VER
@@ -24,7 +24,6 @@
 
 // temporary
 #include <minizinc/prettyprinter.hh>
-//#include <ostream>
 
 #include <map>
 #include <unordered_map>
@@ -382,7 +381,7 @@ private:
     //       {
     //         GCLock lock;
     //         Call* call_EPS_for_LT =
-    //           new Call(Location(),"mzn_float_lt_EPS_coef__", std::vector<Expression*>());
+    //           Call::a(Location(),"mzn_float_lt_EPS_coef__", std::vector<Expression*>());
     //         call_EPS_for_LT->type(Type::parfloat());
     //         call_EPS_for_LT->decl(env.model->matchFn(getEnv()->envi(), call_EPS_for_LT));
     //         float_lt_EPS_coef_ = eval_float(getEnv()->envi(), call_EPS_for_LT);
@@ -588,7 +587,7 @@ private:
       if (ic->removed()) {
         continue;
       }
-      if (Call* c = ic->e()->dynamicCast<Call>()) {
+      if (Call* c = Expression::dynamicCast<Call>(ic->e())) {
         auto ipct = _mCallTypes.find(c->decl());
         if (ipct != _mCallTypes.end()) {
           // No ! here because might be deleted immediately in later versions.
@@ -646,7 +645,7 @@ private:
     if (!fCheckArg) {
       MZN_MIPD_assert_hard(vd->payload() >= 0);
     }
-    if (Id* id = vd->e()->dynamicCast<Id>()) {
+    if (Id* id = Expression::dynamicCast<Id>(vd->e())) {
       //         const int f1 = ( vd->payload()>=0 );
       //         const int f2 = ( id->decl()->payload()>=0 );
       if (!fCheckArg || (id->decl()->payload() >= 0)) {
@@ -665,13 +664,13 @@ private:
         }
         return true;  // in any case
       }
-    } else if (Call* c = vd->e()->dynamicCast<Call>()) {
+    } else if (Call* c = Expression::dynamicCast<Call>(vd->e())) {
       if (lin_exp_int == c->decl() || lin_exp_float == c->decl()) {
         //             std::cerr << "  !E call " << std::flush;
         //             debugprint(c);
         MZN_MIPD_assert_hard(c->argCount() == 3);
         //           ArrayLit* al = c->args()[1]->dynamicCast<ArrayLit>();
-        auto* al = follow_id(c->arg(1))->cast<ArrayLit>();
+        auto* al = Expression::cast<ArrayLit>(follow_id(c->arg(1)));
         MZN_MIPD_assert_hard(al);
         MZN_MIPD_assert_hard(!al->empty());
         if (al->size() == 1) {  // 1-term scalar product in the rhs
@@ -754,12 +753,12 @@ private:
       if (ic->removed()) {
         continue;
       }
-      if (Call* c = ic->e()->dynamicCast<Call>()) {
+      if (Call* c = Expression::dynamicCast<Call>(ic->e())) {
         const bool fIntLinEq = int_lin_eq == c->decl();
         const bool fFloatLinEq = float_lin_eq == c->decl();
         if (fIntLinEq || fFloatLinEq) {
           MZN_MIPD_assert_hard(c->argCount() == 3);
-          auto* al = follow_id(c->arg(1))->cast<ArrayLit>();
+          auto* al = Expression::cast<ArrayLit>(follow_id(c->arg(1)));
           MZN_MIPD_assert_hard(al);
           if (al->size() == 2) {  // 2-term eqn
             LinEq2Vars led;
@@ -783,7 +782,7 @@ private:
             }
           }       /// case with just 1 variable: else if (al->size() == 1) { }
           else {  // larger eqns
-            auto* eVD = get_annotation(c->ann(), Constants::constants().ann.defines_var);
+            auto* eVD = get_annotation(Expression::ann(c), Constants::constants().ann.defines_var);
             if (eVD != nullptr) {
               if (_sCallLinEqN.end() != _sCallLinEqN.find(c)) {
                 continue;
@@ -791,8 +790,7 @@ private:
               _sCallLinEqN.insert(c);  // memorize this call
               DBGOUT_MIPD("       REG N-call ");
               DBGOUT_MIPD_SELF(debugprint(c));
-              Call* pC = eVD->dynamicCast<Call>();
-              MZN_MIPD_assert_hard(pC);
+              Call* pC = Expression::cast<Call>(eVD);
               MZN_MIPD_assert_hard(pC->argCount());
               // Checking all but adding only touched defined vars? Seems too long.
               VarDecl* vd = expr2VarDecl(pC->arg(0));
@@ -845,8 +843,7 @@ private:
   // linexp: z = a^T x+b
   // _lin_eq: a^T x == b
   bool findOrAddDefining(Expression* exp, Call* pC) {
-    Id* pId = exp->dynamicCast<Id>();
-    MZN_MIPD_assert_hard(pId);
+    Id* pId = Expression::cast<Id>(exp);
     VarDecl* vd = pId->decl();
     MZN_MIPD_assert_hard(vd);
     MZN_MIPD_assert_hard(pC->argCount() == 3);
@@ -870,7 +867,7 @@ private:
         nVRest.rhs = -nVRest.rhs;
         ++nVD;
       } else {
-        rhsLin.push_back(std::make_pair(vars[i], coefs[i]));
+        rhsLin.emplace_back(vars[i], coefs[i]);
       }
     }
     MZN_MIPD_assert_hard(1 >= nVD);
@@ -936,7 +933,7 @@ private:
       MZN_MIPD_assert_soft(
           0, "MIPD: STRANGE: registering var connection to itself: " << led << ", skipping");
       MZN_MIPD_ASSERT_FOR_SAT(fabs(led.coefs[0] + led.coefs[1]) < 1e-6,  // TODO param
-                              getEnv()->envi(), led.vd[0]->loc(),
+                              getEnv()->envi(), Expression::loc(led.vd[0]),
                               "Var connection to itself seems to indicate UNSAT: " << led);
       return;
     }
@@ -1185,12 +1182,12 @@ private:
       DBGOUT_MIPD("Clique " << nClique << ": main ref var " << cls.varRef1->id()->str()
                             << ", domain dec: " << sDomain);
 
-      MZN_MIPD_ASSERT_FOR_SAT(!sDomain.empty(), mipd.getEnv()->envi(), cls.varRef1->loc(),
+      MZN_MIPD_ASSERT_FOR_SAT(!sDomain.empty(), mipd.getEnv()->envi(), Expression::loc(cls.varRef1),
                               "clique " << nClique << ": main ref var " << *cls.varRef1->id()
                                         << ", domain decomposition seems empty: " << sDomain);
 
       MZN_MIPD_FLATTENING_ERROR_IF_NOT(sDomain.checkFiniteBounds(), mipd.getEnv()->envi(),
-                                       cls.varRef1->loc(),
+                                       Expression::loc(cls.varRef1),
                                        "variable " << *cls.varRef1->id()
                                                    << " needs finite bounds for linearisation."
                                                       " Or, use indicator constraints. "
@@ -1259,7 +1256,7 @@ private:
           lb = bnds.left;
           ub = bnds.right;
         } else {
-          MZN_MIPD_FLATTENING_ERROR_IF_NOT(0, mipd.getEnv()->envi(), cls.varRef1->loc(),
+          MZN_MIPD_FLATTENING_ERROR_IF_NOT(0, mipd.getEnv()->envi(), Expression::loc(cls.varRef1),
                                            "Variable " << vd->id()->str() << " of type "
                                                        << vd->type().toString(mipd._env->envi())
                                                        << " has a domain.");
@@ -1278,8 +1275,7 @@ private:
       for (Item* pItem : aCalls) {
         auto* pCI = pItem->dynamicCast<ConstraintI>();
         MZN_MIPD_assert_hard(pCI != nullptr);
-        Call* pCall = pCI->e()->dynamicCast<Call>();
-        MZN_MIPD_assert_hard(pCall != nullptr);
+        Call* pCall = Expression::cast<Call>(pCI->e());
         DBGOUT_MIPD_FLUSH("PROPAG CALL  ");
         DBGOUT_MIPD_SELF(debugprint(pCall));
         // check the bounds for bool in reifs?                     TODO
@@ -1414,7 +1410,7 @@ private:
       std::vector<Expression*> pp;
       auto bnds = sDomain.getBounds();
       const long long iMin = mipd.expr2ExprArray(
-          mipd._vVarDescr[cls.varRef1->payload()].pEqEncoding->e()->dynamicCast<Call>()->arg(1),
+          Expression::cast<Call>(mipd._vVarDescr[cls.varRef1->payload()].pEqEncoding->e())->arg(1),
           pp);
       MZN_MIPD_assert_hard(pp.size() >= bnds.right - bnds.left + 1);
       MZN_MIPD_assert_hard(iMin <= bnds.left);
@@ -1430,20 +1426,20 @@ private:
           if (vEE >= static_cast<long long>(iMin + pp.size())) {
             return;
           }
-          if (pp[vEE - iMin]->isa<Id>()) {
-            if (pp[vEE - iMin]->dynamicCast<Id>()->decl()->type().isvar()) {
+          if (Expression::isa<Id>(pp[vEE - iMin])) {
+            if (Expression::type(Expression::cast<Id>(pp[vEE - iMin])->decl()).isvar()) {
               DBGOUT_MIPD_FLUSH(vEE << ", ");
-              setVarDomain(pp[vEE - iMin]->dynamicCast<Id>()->decl(), 0.0, 0.0);
+              setVarDomain(Expression::cast<Id>(pp[vEE - iMin])->decl(), 0.0, 0.0);
             }
           }
         }
         vEE = static_cast<long long>(intv.right + 1);
       }
       for (; vEE < static_cast<long long>(iMin + pp.size()); ++vEE) {
-        if (pp[vEE - iMin]->isa<Id>()) {
-          if (pp[vEE - iMin]->dynamicCast<Id>()->decl()->type().isvar()) {
+        if (Expression::isa<Id>(pp[vEE - iMin])) {
+          if (Expression::type(Expression::cast<Id>(pp[vEE - iMin])->decl()).isvar()) {
             DBGOUT_MIPD_FLUSH(vEE << ", ");
-            setVarDomain(pp[vEE - iMin]->dynamicCast<Id>()->decl(), 0.0, 0.0);
+            setVarDomain(Expression::cast<Id>(pp[vEE - iMin])->decl(), 0.0, 0.0);
           }
         }
       }
@@ -1516,7 +1512,7 @@ private:
         for (Item* pItem : aCalls) {
           auto* pCI = pItem->dynamicCast<ConstraintI>();
           MZN_MIPD_assert_hard(pCI);
-          Call* pCall = pCI->e()->dynamicCast<Call>();
+          Call* pCall = Expression::dynamicCast<Call>(pCI->e());
           MZN_MIPD_assert_hard(pCall);
           DBGOUT_MIPD_FLUSH("IMPL CALL  ");
           DBGOUT_MIPD_SELF(debugprint(pCall));
@@ -1699,7 +1695,8 @@ private:
         std::vector<Expression*> pp;
         auto bnds = sDomain.getBounds();
         const long long iMin = mipd.expr2ExprArray(
-            mipd._vVarDescr[cls.varRef1->payload()].pEqEncoding->e()->dynamicCast<Call>()->arg(1),
+            Expression::cast<Call>(mipd._vVarDescr[cls.varRef1->payload()].pEqEncoding->e())
+                ->arg(1),
             pp);
         MZN_MIPD_assert_hard(pp.size() >= bnds.right - bnds.left + 1);
         MZN_MIPD_assert_hard(iMin <= bnds.left);
@@ -1756,6 +1753,7 @@ private:
         //           if ( 0.0==lb && 0.0==ub ) {
         auto* newDom =
             new BinOp(Location().introduce(), FloatLit::a(lb), BOT_DOTDOT, FloatLit::a(ub));
+        newDom->type(Type::parsetfloat());
         vd->ti()->domain(newDom);
         DBGOUT_MIPD("  NULL OUT:  " << vd->id()->str());
         //           }
@@ -1763,6 +1761,7 @@ private:
         auto* newDom = new SetLit(
             Location().introduce(),
             IntSetVal::a(static_cast<long long int>(lb), static_cast<long long int>(ub)));
+        newDom->type(Type::parsetint());
         //           TypeInst* nti = copy(mipd.getEnv()->envi(),varFlag->ti())->cast<TypeInst>();
         //           nti->domain(newDom);
         vd->ti()->domain(newDom);
@@ -1777,10 +1776,11 @@ private:
       auto* newDom =
           new SetLit(Location().introduce(),
                      IntSetVal::a(static_cast<long long int>(LB), static_cast<long long int>(UB)));
+      newDom->type(Type::parsetint());
       auto* ti = new TypeInst(Location().introduce(), Type::varint(), newDom);
       auto* newVar = new VarDecl(Location().introduce(), ti, mipd.getEnv()->envi().genId());
       newVar->flat(newVar);
-      mipd.getEnv()->envi().flatAddItem(new VarDeclI(Location().introduce(), newVar));
+      mipd.getEnv()->envi().flatAddItem(VarDeclI::a(Location().introduce(), newVar));
       return newVar;
     }
 
@@ -1791,9 +1791,10 @@ private:
       for (auto* v : vars) {
         MZN_MIPD_assert_hard(&v);
         //             throw std::string("addLinConstr: &var=NULL");
-        MZN_MIPD_assert_hard_msg(v->isa<Id>() || v->isa<IntLit>() || v->isa<FloatLit>(),
-                                 "  expression at " << (&v) << " eid = " << v->eid()
-                                                    << " while E_INTLIT=" << Expression::E_INTLIT);
+        MZN_MIPD_assert_hard_msg(
+            Expression::isa<Id>(v) || Expression::isa<IntLit>(v) || Expression::isa<FloatLit>(v),
+            "  expression at " << (&v) << " eid = " << Expression::eid(v)
+                               << " while E_INTLIT=" << Expression::E_INTLIT);
         //             throw std::string("addLinConstr: only id's as variables allowed");
       }
       MZN_MIPD_assert_hard(coefs.size() == vars.size());
@@ -1815,7 +1816,7 @@ private:
       std::vector<Expression*> nx;
       bool fFloat = false;
       for (auto* v : vars) {
-        if (!v->type().isint()) {
+        if (!Expression::type(v).isint()) {
           fFloat = true;
           break;
         }
@@ -1827,11 +1828,11 @@ private:
           if (fabs(coefs[i]) > 1e-8)  /// Only add terms with non-0 coefs. TODO Eps=param
           {
             nc_c.push_back(FloatLit::a(coefs[i]));
-            if (vars[i]->type().isint()) {
+            if (Expression::type(vars[i]).isint()) {
               std::vector<Expression*> i2f_args(1);
               i2f_args[0] = vars[i];
               Call* i2f =
-                  new Call(Location().introduce(), Constants::constants().ids.int2float, i2f_args);
+                  Call::a(Location().introduce(), Constants::constants().ids.int2float, i2f_args);
               i2f->type(Type::varfloat());
               i2f->decl(mipd.getEnv()->model()->matchFn(mipd.getEnv()->envi(), i2f, false));
               EE ret = flat_exp(mipd.getEnv()->envi(), Ctx(), i2f, nullptr,
@@ -1843,11 +1844,11 @@ private:
           }
         }
         args[2] = FloatLit::a(rhs);
-        args[2]->type(Type::parfloat(0));
+        Expression::type(args[2], Type::parfloat(0));
         args[0] = new ArrayLit(Location().introduce(), nc_c);
-        args[0]->type(Type::parfloat(1));
+        Expression::type(args[0], Type::parfloat(1));
         args[1] = new ArrayLit(Location().introduce(), nx);
-        args[1]->type(Type::varfloat(1));
+        Expression::type(args[1], Type::varfloat(1));
         if (CMPT_LE == nCmpType) {
           sName = Constants::constants().ids.float_.lin_le;  // "float_lin_le";
           fDecl = mipd.float_lin_le;
@@ -1861,11 +1862,11 @@ private:
           }
         }
         args[2] = IntLit::a(static_cast<long long int>(rhs));
-        args[2]->type(Type::parint(0));
+        Expression::type(args[2], Type::parint(0));
         args[0] = new ArrayLit(Location().introduce(), nc_c);
-        args[0]->type(Type::parint(1));
+        Expression::type(args[0], Type::parint(1));
         args[1] = new ArrayLit(Location().introduce(), nx);
-        args[1]->type(Type::varint(1));
+        Expression::type(args[1], Type::varint(1));
         if (CMPT_LE == nCmpType) {
           sName = Constants::constants().ids.int_.lin_le;  // "int_lin_le";
           fDecl = mipd.int_lin_le;
@@ -1878,7 +1879,7 @@ private:
         DBGOUT_MIPD_FLUSH(" Found expr ");
         DBGOUT_MIPD_SELF(debugprint(args[0]));
       }
-      auto* nc = new Call(Location().introduce(), ASTString(sName), args);
+      auto* nc = Call::a(Location().introduce(), ASTString(sName), args);
       nc->type(Type::varbool());
       nc->decl(fDecl);
       mipd.getEnv()->envi().flatAddItem(new ConstraintI(Location().introduce(), nc));
@@ -1887,7 +1888,7 @@ private:
     /// domain / reif set of one variable into that for a!her
     void convertIntSet(Expression* e, SetOfIntvReal& s, VarDecl* varTarget, double A, double B) {
       MZN_MIPD_assert_hard(A != 0.0);
-      if (e->type().isIntSet()) {
+      if (Expression::type(e).isIntSet()) {
         IntSetVal* S = eval_intset(mipd.getEnv()->envi(), e);
         IntSetRanges domr(S);
         for (; domr(); ++domr) {  // * A + B
@@ -1903,7 +1904,7 @@ private:
                               : IntvReal::infPlus()));
         }
       } else {
-        assert(e->type().isFloatSet());
+        assert(Expression::type(e).isFloatSet());
         FloatSetVal* S = eval_floatset(mipd.getEnv()->envi(), e);
         FloatSetRanges domr(S);
         for (; domr(); ++domr) {  // * A + B
@@ -1982,7 +1983,7 @@ private:
     //                                  "Expression " << *arg << " is an IntLit!" );
     //       MZN_MIPD_assert_hard( ! arg->dynamicCast<FloatLit>() );
     //       MZN_MIPD_assert_hard( ! arg->dynamicCast<BoolLit>() );
-    Id* id = arg->dynamicCast<Id>();
+    Id* id = Expression::dynamicCast<Id>(arg);
     //       MZN_MIPD_assert_hard(id);
     if (nullptr == id) {
       return nullptr;  // the call using this should be ignored?
@@ -2015,17 +2016,18 @@ private:
   }
 
   static double expr2Const(Expression* arg) {
-    if (auto* il = arg->dynamicCast<IntLit>()) {
-      return (static_cast<double>(il->v().toInt()));
+    if (auto* il = Expression::dynamicCast<IntLit>(arg)) {
+      return (static_cast<double>(IntLit::v(il).toInt()));
     }
-    if (auto* fl = arg->dynamicCast<FloatLit>()) {
-      return (fl->v().toDouble());
+    if (auto* fl = Expression::dynamicCast<FloatLit>(arg)) {
+      return (FloatLit::v(fl).toDouble());
     }
-    if (auto* bl = arg->dynamicCast<BoolLit>()) {
+    if (auto* bl = Expression::dynamicCast<BoolLit>(arg)) {
       return static_cast<double>(bl->v());
     }
     MZN_MIPD_assert_hard_msg(0, "unexpected expression instead of an int/float/bool literal: eid="
-                                    << arg->eid() << " while E_INTLIT=" << Expression::E_INTLIT);
+                                    << Expression::eid(arg)
+                                    << " while E_INTLIT=" << Expression::E_INTLIT);
 
     return 0.0;
   }
@@ -2078,7 +2080,7 @@ private:
     double nSubintvAve = MIPD_stats[N_POSTs_NSubintvSum] / nc;
     MZN_MIPD_assert_hard(MIPD_stats[N_POSTs_NSubintvSum]);
     double dSubSizeAve = MIPD_stats[N_POSTs_SubSizeSum] / MIPD_stats[N_POSTs_NSubintvSum];
-    os << MIPD_stats[N_POSTs_all]
+    os << " " << MIPD_stats[N_POSTs_all]
        << " POSTs"
 #ifdef MZN_MIPDOMAINS_PRINTMORESTATS
           " [ ";
@@ -2105,7 +2107,7 @@ private:
       os << TCliqueSorter::LinEqGraph::dCoefMin << "--" << TCliqueSorter::LinEqGraph::dCoefMax
          << " abs coefs";
     }
-    os << " ... ";
+    os << std::endl;
   }
 
 };  // namespace MiniZinc
