@@ -10,10 +10,13 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include <minizinc/astexception.hh>
+#include <minizinc/aststring.hh>
 #include <minizinc/flatten_internal.hh>
 #include <minizinc/model.hh>
 #include <minizinc/prettyprinter.hh>
+#include <minizinc/type.hh>
 
+#include <algorithm>
 #include <unordered_set>
 #include <vector>
 
@@ -21,30 +24,49 @@
 
 namespace MiniZinc {
 
-Model::FnEntry::FnEntry(FunctionI* fi0)
+Model::FnEntry::FnEntry(EnvI& env, FunctionI* fi0)
     : t(fi0->paramCount()), fi(fi0), isPolymorphic(false), isPolymorphicVariant(false) {
   for (unsigned int i = 0; i < fi->paramCount(); i++) {
     t[i] = fi->param(i)->type();
-    isPolymorphic |= (t[i].bt() == Type::BT_TOP);
+    if (t[i].structBT() && t[i].typeId() == 0) {
+      fi->param(i)->ti()->canonicaliseStruct(env);
+      t[i] = fi->param(i)->ti()->type();
+      fi->param(i)->type(t[i]);
+    }
+    isPolymorphic |= checkPoly(env, t[i]);
+  }
+  if (fi->ti()->type().structBT() && fi->ti()->type().typeId() == 0) {
+    fi->ti()->canonicaliseStruct(env);
   }
 }
 
-bool Model::FnEntry::operator<(const Model::FnEntry& f) const {
-  assert(!compare(*this, f) || !compare(f, *this));
-  return compare(*this, f);
+bool Model::FnEntry::checkPoly(const EnvI& env, const Type& t) {
+  if (t.bt() == Type::BT_TOP) {
+    return true;
+  }
+  if (t.structBT()) {
+    StructType* st = env.getStructType(t);
+    for (size_t i = 0; i < st->size(); ++i) {
+      if (checkPoly(env, (*st)[i])) {
+        return true;
+      }
+    }
+  }
+  return false;
 }
 
-bool Model::FnEntry::compare(const Model::FnEntry& e1, const Model::FnEntry& e2) {
+bool Model::FnEntry::compare(const EnvI& env, const Model::FnEntry& e1, const Model::FnEntry& e2) {
   if (e1.t.size() < e2.t.size()) {
     return true;
   }
   if (e1.t.size() == e2.t.size()) {
     for (unsigned int i = 0; i < e1.t.size(); i++) {
       if (e1.t[i] != e2.t[i]) {
-        if (e1.t[i].isSubtypeOf(e2.t[i], true)) {
+        if (e1.t[i].isSubtypeOf(env, e2.t[i], true)) {
+          assert(!e2.t[i].isSubtypeOf(env, e1.t[i], true));
           return true;
         }
-        if (e2.t[i].isSubtypeOf(e1.t[i], true)) {
+        if (e2.t[i].isSubtypeOf(env, e1.t[i], true)) {
           return false;
         }
         switch (e1.t[i].cmp(e2.t[i])) {
@@ -119,6 +141,24 @@ void Model::setOutputItem(OutputI* oi) {
 namespace {
 
 enum PossibleBaseTypes { PBT_ANY, PBT_ALL, PBT_BIFS, PBT_BIF, PBT_I };
+PossibleBaseTypes pbt_from_type(const Type& t) {
+  if (t.any()) {
+    return PBT_ANY;
+  }
+  if (t.isvar()) {
+    if (t.isOpt()) {
+      return PBT_BIF;
+    }
+    if (t.isSet()) {
+      return PBT_I;
+    }
+    return PBT_BIFS;
+  }
+  if (t.isSet()) {
+    return PBT_BIF;
+  }
+  return PBT_ALL;
+}
 PossibleBaseTypes pbt_join(PossibleBaseTypes pbt0, PossibleBaseTypes pbt1) {
   if (pbt0 == PBT_ANY) {
     return pbt1;
@@ -248,9 +288,72 @@ void increment_type(Type& t, PossibleBaseTypes pbt) {
 
 }  // namespace
 
-void Model::addPolymorphicInstances(Model::FnEntry& fe, std::vector<FnEntry>& entries) {
-  entries.push_back(fe);
+// For each TIId, store pointers to the type objects that refer to this TIId
+struct TIIDInfo {
+  std::vector<Type*> t;
+  PossibleBaseTypes pbt;
+  TIIDInfo(std::vector<Type*> t0, PossibleBaseTypes pbt0) : t(std::move(t0)), pbt(pbt0) {}
+};
+
+void TypeInst::collectTypeIds(std::unordered_map<ASTString, size_t>& seen_tiids,
+                              std::vector<TIIDInfo>& type_ids) const {
+  auto* al = Expression::cast<ArrayLit>(domain());
+  for (size_t i = 0; i < al->size(); i++) {
+    auto* ti = Expression::cast<TypeInst>((*al)[i]);
+    if (ti->type().bt() == Type::BT_TOP) {
+      // If type is top, either this is a TIId, or it is caused by <>
+      if (ti->domain() != nullptr) {
+        assert(Expression::isa<TIId>(ti->domain()));
+        TIId* id0 = Expression::cast<TIId>(ti->domain());
+        auto it = seen_tiids.find(id0->v());
+        if (it == seen_tiids.end()) {
+          type_ids.emplace_back(
+              std::vector<Type*>({&ti->_type}),
+              pbt_join(ti->type().any() ? PBT_ANY : PBT_ALL, pbt_from_type(ti->type())));
+          seen_tiids.emplace(id0->v(), type_ids.size() - 1);
+        } else {
+          TIIDInfo& info = type_ids[it->second];
+          info.pbt = pbt_join(info.pbt, pbt_from_type(ti->type()));
+          info.t.push_back(&ti->_type);
+        }
+      }
+    } else if (ti->type().structBT()) {
+      size_t size = seen_tiids.size();
+      ti->collectTypeIds(seen_tiids, type_ids);
+    }
+  }
+}
+
+void Model::addPolymorphicInstances(EnvI& env, Model::FnEntry& fe, std::vector<FnEntry>& entries) {
+  auto addEntry = [&](Model::FnEntry& toAdd) {
+    for (auto& entry : entries) {
+      if (entry.t == toAdd.t) {
+        bool more_specific = true;
+        for (unsigned int i = 0; i < toAdd.fi->paramCount(); i++) {
+          // If all parameters of the entry we are adding are subtypes of an
+          // existing function, this one should take priority
+          //
+          // E.g. (bool, int) is preferred over ($T, $U)
+          auto oldParamType = entry.fi->param(i)->type();
+          auto newParamType = toAdd.fi->param(i)->type();
+          if (!newParamType.isSubtypeOf(env, oldParamType, false)) {
+            more_specific = false;
+            break;
+          }
+        }
+        if (more_specific) {
+          entry = toAdd;
+        }
+        return;
+      }
+    }
+    // Entry not yet added
+    entries.push_back(toAdd);
+  };
+
+  addEntry(fe);
   if (fe.isPolymorphic) {
+    GCLock lock;
     FnEntry cur = fe;
     cur.isPolymorphicVariant = true;
 
@@ -273,57 +376,13 @@ void Model::addPolymorphicInstances(Model::FnEntry& fe, std::vector<FnEntry>& en
 
      */
 
-    // For each TIId, store pointers to the type objects that refer to this TIId
-    struct TIIDInfo {
-      std::vector<Type*> t;
-      PossibleBaseTypes pbt;
-      TIIDInfo(std::vector<Type*> t0, PossibleBaseTypes pbt0) : t(std::move(t0)), pbt(pbt0) {}
-    };
+    std::unordered_map<ASTString, size_t> type_id_map;
     std::vector<TIIDInfo> type_ids;
-    std::unordered_set<ASTString> seen_tiids;
 
-    // First step: initialise all type variables to bool
-    // and collect them in the type_ids vector
-    for (unsigned int i = 0; i < cur.t.size(); i++) {
-      if (cur.t[i].bt() == Type::BT_TOP) {
-        assert(cur.fi->param(i)->ti()->domain() && cur.fi->param(i)->ti()->domain()->isa<TIId>());
-        TIId* id0 = cur.fi->param(i)->ti()->domain()->cast<TIId>();
-        if (seen_tiids.find(id0->v()) == seen_tiids.end()) {
-          seen_tiids.insert(id0->v());
-          // New tiid, collect all occurrences
-          std::vector<Type*> t;
-          PossibleBaseTypes pbt = cur.fi->param(i)->ti()->type().any() ? PBT_ANY : PBT_ALL;
-          for (unsigned int j = i; j < cur.t.size(); j++) {
-            if ((cur.fi->param(j)->ti()->domain() != nullptr) &&
-                cur.fi->param(j)->ti()->domain()->isa<TIId>()) {
-              TIId* id1 = cur.fi->param(j)->ti()->domain()->cast<TIId>();
-              if (id0->v() == id1->v()) {
-                // Found parameter with same type variable
-                if (cur.fi->param(j)->ti()->type().any()) {
-                  pbt = pbt_join(pbt, PBT_ANY);
-                } else if (cur.fi->param(j)->type().isvar()) {
-                  if (cur.fi->param(j)->type().isOpt()) {
-                    pbt = pbt_join(pbt, PBT_BIF);
-                  } else if (cur.fi->param(j)->type().isSet()) {
-                    pbt = pbt_join(pbt, PBT_I);
-                  } else {
-                    pbt = pbt_join(pbt, PBT_BIFS);
-                  }
-                } else {
-                  if (cur.fi->param(j)->type().isSet()) {
-                    pbt = pbt_join(pbt, PBT_BIF);
-                  } else {
-                    pbt = pbt_join(pbt, PBT_ALL);
-                  }
-                }
-                t.push_back(&cur.t[j]);
-              }
-            }
-          }
-          type_ids.emplace_back(t, pbt);
-        }
-      }
-    }
+    // Create a parameter TypeInst in the format of a tuple TypeInst
+    // Immediately copy so the internal TypeInst values can be changed
+    auto* paramtuple = Expression::cast<TypeInst>(copy(env, cur.fi->paramTypes()));
+    paramtuple->collectTypeIds(type_id_map, type_ids);
 
     std::vector<size_t> stack;
     for (size_t i = 0; i < type_ids.size(); i++) {
@@ -336,17 +395,20 @@ void Model::addPolymorphicInstances(Model::FnEntry& fe, std::vector<FnEntry>& en
 
     while (!stack.empty()) {
       if (stack.back() == final_id) {
-        // If this instance isn't in entries yet, add it
-        bool alreadyDefined = false;
-        for (auto& entry : entries) {
-          if (entry.t == cur.t) {
-            alreadyDefined = true;
-            break;
+        // New complete instance
+        // First, update cur types
+        auto* tis = Expression::cast<ArrayLit>(paramtuple->domain());
+        for (size_t i = 0; i < tis->size(); ++i) {
+          cur.t[i] = Expression::type((*tis)[i]);
+          if (cur.t[i].bt() == Type::BT_TUPLE && cur.t[i].typeId() == 0) {
+            env.registerTupleType(Expression::cast<TypeInst>((*tis)[i]));
+            cur.t[i] = Expression::type((*tis)[i]);
+          } else if (cur.t[i].bt() == Type::BT_RECORD && cur.t[i].typeId() == 0) {
+            env.registerRecordType(Expression::cast<TypeInst>((*tis)[i]));
+            cur.t[i] = Expression::type((*tis)[i]);
           }
         }
-        if (!alreadyDefined) {
-          entries.push_back(cur);
-        }
+        addEntry(cur);
       }
 
       const Type& back_t = *type_ids[stack.back()].t[0];
@@ -378,12 +440,13 @@ bool Model::registerFn(EnvI& env, FunctionI* fi, bool keepSorted, bool throwIfDu
   if (i_id == m->_fnmap.end()) {
     // new element
     std::vector<FnEntry> v;
-    FnEntry fe(fi);
-    addPolymorphicInstances(fe, v);
+    FnEntry fe(env, fi);
+    addPolymorphicInstances(env, fe, v);
     m->_fnmap.insert(std::pair<ASTString, std::vector<FnEntry> >(fi->id(), v));
   } else {
     // add to list of existing elements
     std::vector<FnEntry>& v = i_id->second;
+    FnEntry fe(env, fi);  // Create now so that struct types get canonicalised
     for (auto& i : v) {
       if (i.fi == fi) {
         return true;
@@ -394,13 +457,11 @@ bool Model::registerFn(EnvI& env, FunctionI* fi, bool keepSorted, bool throwIfDu
         for (unsigned int j = 0; j < fi->paramCount(); j++) {
           Type t1 = i.fi->param(j)->type();
           Type t2 = fi->param(j)->type();
-          t1.enumId(0);
-          t2.enumId(0);
           if (t1 != t2) {
             alleq = false;
           }
-          t1.ti(Type::TI_PAR);
-          t2.ti(Type::TI_PAR);
+          t1.mkPar(env);
+          t2.mkPar(env);
           if (t1 != t2) {
             eqExceptInst = false;
           }
@@ -418,7 +479,17 @@ bool Model::registerFn(EnvI& env, FunctionI* fi, bool keepSorted, bool throwIfDu
             if (Call* deprecated = i.fi->ann().getCall(env.constants.ann.mzn_deprecated)) {
               fi->ann().add(deprecated);
             }
-            i = fi;
+            FunctionI* old_fi = i.fi;
+            i = FnEntry(env, fi);
+            // If we are replacing a polymorphic function using a new polymorphic function, then
+            // replace in all entries generated using addPolymorphicInstances
+            if (i.isPolymorphic) {
+              for (auto& j : v) {
+                if (j.fi == old_fi) {
+                  j.fi = fi;
+                }
+              }
+            }
           } else if (Call* deprecated = fi->ann().getCall(env.constants.ann.mzn_deprecated)) {
             i.fi->ann().add(deprecated);
           }
@@ -427,12 +498,16 @@ bool Model::registerFn(EnvI& env, FunctionI* fi, bool keepSorted, bool throwIfDu
         if (eqExceptInst) {
           Type t1 = i.fi->ti()->type();
           Type t2 = fi->ti()->type();
-          t1.enumId(0);
-          t2.enumId(0);
-          t1.ti(Type::TI_PAR);
-          t2.ti(Type::TI_PAR);
-          t1.ot(Type::OT_PRESENT);
-          t2.ot(Type::OT_PRESENT);
+          t1.mkPar(env);
+          t2.mkPar(env);
+          t1.mkPresent(env);
+          t2.mkPresent(env);
+          if (t1.bt() == Type::BT_INT) {
+            t1.typeId(0);
+          }
+          if (t2.bt() == Type::BT_INT) {
+            t2.typeId(0);
+          }
           if (t1 != t2) {
             throw TypeError(env, fi->loc(),
                             "function with same type up to par/var but different return type "
@@ -442,10 +517,11 @@ bool Model::registerFn(EnvI& env, FunctionI* fi, bool keepSorted, bool throwIfDu
         }
       }
     }
-    FnEntry fe(fi);
-    addPolymorphicInstances(fe, v);
+    addPolymorphicInstances(env, fe, v);
     if (keepSorted) {
-      std::sort(i_id->second.begin(), i_id->second.end());
+      std::sort(v.begin(), v.end(), [&env](const Model::FnEntry& e1, const Model::FnEntry& e2) {
+        return Model::FnEntry::compare(env, e1, e2);
+      });
     }
   }
   if (fi->id() == env.constants.ids.mzn_reverse_map_var) {
@@ -476,15 +552,17 @@ void compute_possible_matches(EnvI& env, const Model* m, const ASTString& ident,
                               std::vector<FunctionI*>& ret) {
   // Go through the types in this order: var opt, var, par opt, par
   std::vector<Type> vp(ta.size());
-  for (unsigned int i = 0; i < ta.size(); i++) {
-    vp[i] = ta[i];
-    vp[i].ti(Type::TI_VAR);
-    if (vp[i].st() == Type::ST_PLAIN) {
-      vp[i].ot(Type::OT_OPTIONAL);
+  auto reset_types_from = [&](size_t from) {
+    for (size_t i = from; i < ta.size(); i++) {
+      vp[i] = ta[i];
+      vp[i].mkVar(env);
+      if (vp[i].st() == Type::ST_PLAIN) {
+        vp[i].mkOpt(env);
+      }
     }
-  }
+  };
+  reset_types_from(0);
   int finalType = static_cast<int>(ta.size()) - 1;
-  std::vector<int> stack;
 
   for (;;) {
     auto* fi = m->matchFn(env, ident, vp, false);
@@ -497,29 +575,8 @@ void compute_possible_matches(EnvI& env, const Model* m, const ASTString& ident,
     int i = finalType;
     for (; i >= 0; i--) {
       Type& t = vp[i];
-      if (t.ti() != Type::TI_PAR || t.ot() != Type::OT_PRESENT) {
-        if (t.ti() == Type::TI_VAR) {
-          if (t.ot() == Type::OT_OPTIONAL) {
-            // var opt, turn into var
-            t.ot(Type::OT_PRESENT);
-          } else {
-            // var, turn into par opt
-            if (t.st() == Type::ST_PLAIN) {
-              t.ot(Type::OT_OPTIONAL);
-            }
-            t.ti(Type::TI_PAR);
-          }
-        } else {
-          // this is par opt, turn into par
-          t.ot(Type::OT_PRESENT);
-        }
-        for (unsigned int j = i + 1; j < ta.size(); j++) {
-          vp[j] = ta[j];
-          vp[j].ti(Type::TI_VAR);
-          if (vp[j].st() == Type::ST_PLAIN) {
-            vp[j].ot(Type::OT_OPTIONAL);
-          }
-        }
+      if (t.decrement(env)) {
+        reset_types_from(i + 1);
         break;
       }
     }
@@ -546,14 +603,14 @@ std::vector<FunctionI*> Model::possibleMatches(EnvI& env, const ASTString& ident
   // Try reified/non-reified versions
   if (ident.endsWith("_reif")) {
     std::string ident_s(ident.c_str());
-    ASTString baseIdent = ident_s.substr(0, ident_s.length() - 5);
+    ASTString baseIdent(ident_s.substr(0, ident_s.length() - 5));
     std::vector<Type> ta_b = ta;
     compute_possible_matches(env, this, EnvI::halfReifyId(baseIdent), ta_b, matched, ret);
     ta_b.pop_back();
     compute_possible_matches(env, this, baseIdent, ta_b, matched, ret);
   } else if (ident.endsWith("_imp")) {
     std::string ident_s(ident.c_str());
-    ASTString baseIdent = ident_s.substr(0, ident_s.length() - 4);
+    ASTString baseIdent(ident_s.substr(0, ident_s.length() - 4));
     std::vector<Type> ta_b = ta;
     compute_possible_matches(env, this, env.reifyId(baseIdent), ta_b, matched, ret);
     ta_b.pop_back();
@@ -614,17 +671,20 @@ void Model::mergeStdLib(EnvI& env, Model* m) const {
       }
     }
   }
-  m->sortFn();
+  m->sortFn(env);
 }
 
-void Model::sortFn() {
+void Model::sortFn(const EnvI& env) {
   Model* m = this;
   while (m->_parent != nullptr) {
     m = m->_parent;
   }
   for (auto& it : m->_fnmap) {
     // Sort all functions by type
-    std::sort(it.second.begin(), it.second.end());
+    std::sort(it.second.begin(), it.second.end(),
+              [&env](const Model::FnEntry& e1, const Model::FnEntry& e2) {
+                return Model::FnEntry::compare(env, e1, e2);
+              });
   }
 }
 
@@ -636,7 +696,7 @@ void Model::fixFnMap() {
   for (auto& it : m->_fnmap) {
     for (auto& i : it.second) {
       for (unsigned int j = 0; j < i.t.size(); j++) {
-        if (i.t[j].isunknown()) {
+        if (i.t[j].isunknown() || i.t[j].structBT()) {
           i.t[j] = i.fi->param(j)->type();
         }
       }
@@ -656,9 +716,9 @@ void Model::checkFnValid(EnvI& env, std::vector<TypeError>& errors) {
     for (const auto& fe : variants) {
       FunctionI* fi = fe.fi;
       Type ret = fi->ti()->type();
-      if (fi->e() != nullptr || ret.isAnn() || fi->builtins.e != nullptr ||
-          fi->builtins.b != nullptr || fi->builtins.f != nullptr || fi->builtins.i != nullptr ||
-          fi->builtins.s != nullptr || fi->builtins.fs != nullptr || fi->builtins.str != nullptr ||
+      if (fi->e() != nullptr || fi->builtins.e != nullptr || fi->builtins.b != nullptr ||
+          fi->builtins.f != nullptr || fi->builtins.i != nullptr || fi->builtins.s != nullptr ||
+          fi->builtins.fs != nullptr || fi->builtins.str != nullptr ||
           fi->ann().contains(env.constants.ann.mzn_internal_representation)) {
         continue;
       }
@@ -666,10 +726,10 @@ void Model::checkFnValid(EnvI& env, std::vector<TypeError>& errors) {
         std::vector<Type> tys(fi->paramCount());
         for (int i = 0; i < fi->paramCount(); ++i) {
           tys[i] = fi->param(i)->type();
-          tys[i].ti(Type::TI_PAR);
+          tys[i].mkPar(env);
         }
         fi = matchFn(env, fi->id(), tys, true);
-        ret.ti(Type::TI_PAR);
+        ret.mkPar(env);
         if (fi != nullptr && ret == fi->ti()->type()) {
           continue;
         }
@@ -678,12 +738,12 @@ void Model::checkFnValid(EnvI& env, std::vector<TypeError>& errors) {
         continue;
       }
       // Par functions that are not implemented in the compiler should have an implementation
-      if (ret.isPar()) {
+      if (!ret.isAnn() && ret.isPar()) {
         errors.emplace_back(env, fi->loc(),
                             "Parameter type function is missing its implementation");
         continue;
       }
-      if (!ret.isvarbool()) {
+      if (!ret.isAnn() && !ret.isvarbool()) {
         errors.emplace_back(
             env, fi->loc(),
             "FlatZinc builtin functions must be predicates (i.e., have `var bool` return type)");
@@ -691,12 +751,19 @@ void Model::checkFnValid(EnvI& env, std::vector<TypeError>& errors) {
       }
       for (int i = 0; i < fi->paramCount(); ++i) {
         const Type& t = fi->param(i)->type();
-        if (t.isOpt() || t.bt() == Type::BT_TOP) {
+        if (t.isOpt() || t.structBT() || t.bt() == Type::BT_TOP) {
           errors.emplace_back(
-              env, fi->param(i)->loc(),
+              env, Expression::loc(fi->param(i)),
               "FlatZinc builtins are not allowed to have arguments of type " + t.toString(env));
           break;  // Break from parameter, but does continue in FnMap
         }
+        if (t.dim() > 1) {
+          errors.emplace_back(env, Expression::loc(fi->param(i)),
+                              "Type " + t.toString(env) +
+                                  " is not allowed in as a FlatZinc builtin argument, arrays must "
+                                  "be one dimensional");
+          break;  // Break from parameter, but does continue in FnMap
+        }
       }
     }
   }
@@ -720,8 +787,12 @@ void Model::checkFnOverloading(EnvI& env) {
         for (unsigned int i = 0; i < cur->paramCount(); i++) {
           Type t1 = cur->param(i)->type();
           Type t2 = cmp->param(i)->type();
-          t1.enumId(0);
-          t2.enumId(0);
+          if (t1.bt() == Type::BT_INT) {
+            t1.typeId(0);
+          }
+          if (t2.bt() == Type::BT_INT) {
+            t2.typeId(0);
+          }
           if (t1 != t2) {
             allEqual = false;
             break;
@@ -751,15 +822,15 @@ int match_idx(std::vector<FunctionI*>& matched, Expression*& botarg, EnvI& env,
     if (fi_t.size() == args.size()) {
       bool match = true;
       for (unsigned int j = 0; j < args.size(); j++) {
-        if (!env.isSubtype(args[j]->type(), fi_t[j], strictEnums)) {
+        if (!env.isSubtype(Expression::type(args[j]), fi_t[j], strictEnums)) {
 #ifdef MZN_DEBUG_FUNCTION_REGISTRY
-          std::cerr << args[j]->type().toString(env) << " does not match " << fi_t[j].toString(env)
-                    << "\n";
+          std::cerr << Expression::type(args[j]).toString(env) << " does not match "
+                    << fi_t[j].toString(env) << "\n";
 #endif
           match = false;
           break;
         }
-        if (args[j]->type().isbot() && fi_t[j].bt() != Type::BT_TOP) {
+        if (Expression::type(args[j]).isbot() && fi_t[j].bt() != Type::BT_TOP) {
           botarg = args[j];
         }
       }
@@ -775,6 +846,46 @@ int match_idx(std::vector<FunctionI*>& matched, Expression*& botarg, EnvI& env,
 }
 }  // namespace
 
+FunctionI* Model::matchReification(EnvI& env, const ASTString& id,
+                                   const std::vector<Expression*>& args, bool canHalfReify,
+                                   bool strictEnums) const {
+  std::vector<Type> t;
+  t.reserve(args.size());
+  for (const auto* e : args) {
+    t.push_back(Expression::type(e));
+  }
+  return this->matchReification(env, id, t, canHalfReify, strictEnums);
+}
+
+FunctionI* Model::matchReification(EnvI& env, const ASTString& id, const std::vector<Type>& t,
+                                   bool canHalfReify, bool strictEnums) const {
+  ASTString reif_id = env.reifyId(id);
+  FunctionI* reif_decl = this->matchFn(env, reif_id, t, strictEnums);
+  if (canHalfReify) {
+    ASTString imp_id = EnvI::halfReifyId(id);
+    if (FunctionI* imp_decl = this->matchFn(env, imp_id, t, strictEnums)) {
+      // If reification failed, then go with half-reification immediately
+      if (reif_decl == nullptr) {
+        return imp_decl;
+      }
+
+      // If there is both a reification and a half-reification, then make sure the half-reification
+      // is at least as specific as the reification. That is the parameters of the matching
+      // half-reification declaration should be the same or subtypes of the matching reification.
+      assert(imp_decl->paramCount() == reif_decl->paramCount());
+      for (size_t i = 0; i < imp_decl->paramCount(); ++i) {
+        Type a = imp_decl->param(i)->ti()->type();
+        Type b = reif_decl->param(i)->ti()->type();
+        if (!env.isSubtype(a, b, strictEnums)) {
+          return reif_decl;
+        }
+      }
+      return imp_decl;
+    }
+  }
+  return reif_decl;
+}
+
 FunctionI* Model::matchFn(EnvI& env, const ASTString& id, const std::vector<Expression*>& args,
                           bool strictEnums) const {
   if (id == env.constants.varRedef->id()) {
@@ -799,10 +910,11 @@ FunctionI* Model::matchFn(EnvI& env, const ASTString& id, const std::vector<Expr
     return matched[0];
   }
   Type t = matched[0]->ti()->type();
-  t.ti(Type::TI_PAR);
+  t.mkPar(env);
   for (unsigned int i = 1; i < matched.size(); i++) {
     if (!env.isSubtype(t, matched[i]->ti()->type(), strictEnums)) {
-      throw TypeError(env, botarg->loc(), "ambiguous overloading on return type of function");
+      throw TypeError(env, Expression::loc(botarg),
+                      "ambiguous overloading on return type of function");
     }
   }
   return matched[0];
@@ -838,7 +950,7 @@ FunctionI* Model::matchFn(EnvI& env, Call* c, bool strictEnums, bool throwIfNotF
       if (!mostSimilar.empty()) {
         oss << ", did you mean `" << mostSimilar << "'?";
       }
-      throw TypeError(env, c->loc(), oss.str());
+      throw TypeError(env, Expression::loc(c), oss.str());
     }
     return nullptr;
   }
@@ -853,16 +965,16 @@ FunctionI* Model::matchFn(EnvI& env, Call* c, bool strictEnums, bool throwIfNotF
     if (fi_t.size() == c->argCount()) {
       bool match = true;
       for (unsigned int j = 0; j < c->argCount(); j++) {
-        if (!env.isSubtype(c->arg(j)->type(), fi_t[j], strictEnums)) {
+        if (!env.isSubtype(Expression::type(c->arg(j)), fi_t[j], strictEnums)) {
 #ifdef MZN_DEBUG_FUNCTION_REGISTRY
-          std::cerr << c->arg(j)->type().toString(env) << " does not match "
+          std::cerr << Expression::type(c->arg(j)).toString(env) << " does not match "
                     << fi_t[j].toString(env) << "\n";
           std::cerr << "Wrong argument is " << *c->arg(j) << "\n";
 #endif
           match = false;
           break;
         }
-        if (c->arg(j)->type().isbot() && fi_t[j].bt() != Type::BT_TOP) {
+        if (Expression::type(c->arg(j)).isbot() && fi_t[j].bt() != Type::BT_TOP) {
           botarg = c->arg(j);
         }
       }
@@ -881,7 +993,7 @@ FunctionI* Model::matchFn(EnvI& env, Call* c, bool strictEnums, bool throwIfNotF
       oss << "no function or predicate with this signature found: `";
       oss << c->id() << "(";
       for (unsigned int i = 0; i < c->argCount(); i++) {
-        oss << c->arg(i)->type().toString(env);
+        oss << Expression::type(c->arg(i)).toString(env);
         if (i < c->argCount() - 1) {
           oss << ",";
         }
@@ -900,9 +1012,9 @@ FunctionI* Model::matchFn(EnvI& env, Call* c, bool strictEnums, bool throwIfNotF
         i.fi->e(body);
         if (fi_t.size() == c->argCount()) {
           for (unsigned int j = 0; j < c->argCount(); j++) {
-            if (!env.isSubtype(c->arg(j)->type(), fi_t[j], strictEnums)) {
+            if (!env.isSubtype(Expression::type(c->arg(j)), fi_t[j], strictEnums)) {
               oss << "    (argument " << (j + 1) << " expects type " << fi_t[j].toString(env);
-              oss << ", but type " << c->arg(j)->type().toString(env) << " given)\n";
+              oss << ", but type " << Expression::type(c->arg(j)).toString(env) << " given)\n";
             }
           }
         } else {
@@ -910,7 +1022,7 @@ FunctionI* Model::matchFn(EnvI& env, Call* c, bool strictEnums, bool throwIfNotF
               << (i.fi->paramCount() == 1 ? "" : "s") << ", but " << c->argCount() << " given)\n";
         }
       }
-      throw TypeError(env, c->loc(), oss.str());
+      throw TypeError(env, Expression::loc(c), oss.str());
     }
     return nullptr;
   }
@@ -918,10 +1030,11 @@ FunctionI* Model::matchFn(EnvI& env, Call* c, bool strictEnums, bool throwIfNotF
     return matched[0];
   }
   Type t = matched[0]->ti()->type();
-  t.ti(Type::TI_PAR);
+  t.mkPar(env);
   for (unsigned int i = 1; i < matched.size(); i++) {
     if (!env.isSubtype(t, matched[i]->ti()->type(), strictEnums)) {
-      throw TypeError(env, botarg->loc(), "ambiguous overloading on return type of function");
+      throw TypeError(env, Expression::loc(botarg),
+                      "ambiguous overloading on return type of function");
     }
   }
   return matched[0];
@@ -952,6 +1065,9 @@ int first_overloaded(EnvI& env, const std::vector<Model::FnEntry>& v_f, int i_f)
 
 bool Model::sameOverloading(EnvI& env, const std::vector<Expression*>& args, FunctionI* f,
                             FunctionI* g) const {
+  if (f->id() == env.constants.varRedef->id() || g->id() == env.constants.varRedef->id()) {
+    return false;
+  }
   const Model* m = this;
   while (m->_parent != nullptr) {
     m = m->_parent;
@@ -996,7 +1112,9 @@ FunctionI* Model::matchRevMap(EnvI& env, const Type& t0) const {
     m = m->_parent;
   }
   Type t = t0;
-  t.enumId(0);
+  if (t.bt() == Type::BT_INT) {
+    t.typeId(0);
+  }
   auto it = _revmapmap.find(t.toInt());
   if (it != _revmapmap.end()) {
     return it->second;
