@@ -297,8 +297,9 @@ void unify(EnvI& env, std::vector<VarDecl*>& deletedVarDecls, Id* id0, Id* id1)
       assert(env.outputFlatVarOccurrences.find(id1->decl()) != -1);
       VarDecl* id1_output =
           (*env.output)[env.outputFlatVarOccurrences.find(id1->decl())]->cast<VarDeclI>()->e();
-      if (id0_output->e() == nullptr) {
-        id0_output->e(id1_output->id());
+      auto* decl = Expression::cast<VarDecl>(follow_id_to_decl(id0_output));
+      if (decl->e() == nullptr) {
+        decl->e(id1_output->id());
       }
     }
 
@@ -311,7 +312,8 @@ void simplify_bool_constraint(EnvI& env, Item* ii, VarDecl* vd, bool& remove,
                               std::deque<unsigned int>& vardeclQueue,
                               std::deque<Item*>& constraintQueue, std::vector<Item*>& toRemove,
                               std::vector<VarDecl*>& deletedVarDecls,
-                              std::unordered_map<Expression*, int>& nonFixedLiteralCount);
+                              std::unordered_map<Expression*, int>& nonFixedLiteralCount,
+                              std::vector<std::pair<Id*, Item*>>& toUnsubscribe);
 
 bool simplify_constraint(EnvI& env, Item* ii, std::vector<VarDecl*>& deletedVarDecls,
                          std::deque<Item*>& constraintQueue,
@@ -880,6 +882,7 @@ void optimize(Env& env, bool chain_compression) {
 
           // Handle all boolean constraints that involve this variable
           if (it.first) {
+            std::vector<std::pair<Id*, Item*>> toUnsubscribe;
             for (auto* item : *it.second) {
               if (item->removed()) {
                 continue;
@@ -894,7 +897,14 @@ void optimize(Env& env, bool chain_compression) {
               }
               // Simplify the constraint *item (which depends on this variable)
               simplify_bool_constraint(envi, item, vd, remove, vardeclQueue, constraintQueue,
-                                       toRemove, deletedVarDecls, nonFixedLiteralCount);
+                                       toRemove, deletedVarDecls, nonFixedLiteralCount,
+                                       toUnsubscribe);
+            }
+            for (auto& unsubscribe : toUnsubscribe) {
+              auto it = envi.varOccurrences.itemMap.find(unsubscribe.first);
+              if (it.first) {
+                it.second->erase(unsubscribe.second);
+              }
             }
           }
           // Actually remove all items that have become unnecessary in the step above
@@ -1743,7 +1753,8 @@ int bool_state(EnvI& env, Expression* e) {
   return static_cast<int>(id->decl()->ti()->domain() == env.constants.literalTrue);
 }
 
-int decrement_non_fixed_vars(std::unordered_map<Expression*, int>& nonFixedLiteralCount, Call* c) {
+int decrement_non_fixed_vars(EnvI& env, Item* ii, std::vector<std::pair<Id*, Item*>>& toUnsubscribe,
+                             std::unordered_map<Expression*, int>& nonFixedLiteralCount, Call* c) {
   auto it = nonFixedLiteralCount.find(c);
   if (it == nonFixedLiteralCount.end()) {
     int nonFixedVars = 0;
@@ -1755,6 +1766,9 @@ int decrement_non_fixed_vars(std::unordered_map<Expression*, int>& nonFixedLiter
             (Expression::isa<Id>((*al)[j]) &&
              Expression::cast<Id>((*al)[j])->decl()->ti()->domain() != nullptr)) {
           nonFixedVars--;
+          if (Expression::isa<Id>((*al)[j])) {
+            toUnsubscribe.emplace_back(Expression::cast<Id>((*al)[j]), ii);
+          }
         }
       }
     }
@@ -1769,7 +1783,8 @@ void simplify_bool_constraint(EnvI& env, Item* ii, VarDecl* vd, bool& remove,
                               std::deque<unsigned int>& vardeclQueue,
                               std::deque<Item*>& constraintQueue, std::vector<Item*>& toRemove,
                               std::vector<VarDecl*>& deletedVarDecls,
-                              std::unordered_map<Expression*, int>& nonFixedLiteralCount) {
+                              std::unordered_map<Expression*, int>& nonFixedLiteralCount,
+                              std::vector<std::pair<Id*, Item*>>& toUnsubscribe) {
   if (ii->isa<SolveI>()) {
     remove = false;
     return;
@@ -1929,7 +1944,7 @@ void simplify_bool_constraint(EnvI& env, Item* ii, VarDecl* vd, bool& remove,
         }
       }
     } else {
-      int nonfixed = decrement_non_fixed_vars(nonFixedLiteralCount, c);
+      int nonfixed = decrement_non_fixed_vars(env, ii, toUnsubscribe, nonFixedLiteralCount, c);
       bool isConjunction = (c->id() == env.constants.ids.forall);
       assert(nonfixed >= 0);
       if (nonfixed <= 1) {
