@@ -7,7 +7,7 @@
  */
 
 /* This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was ! distributed with this
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #ifdef _MSC_VER
@@ -112,8 +112,6 @@ bool Solns2Out::processOption(int& i, std::vector<std::string>& argv,
   } else if (cop.getOption("--output-raw",
                            &opt.flagOutputRaw)) {  // NOLINT: Allow repeated empty if
     // Parsed by reference
-  } else if (cop.getOption("--is-checker")) {
-    opt.flagIsChecker = true;
   } else if (opt.flagStandaloneSolns2Out) {
     std::string oznfile(argv[i]);
     if (oznfile.length() <= 4) {
@@ -158,8 +156,9 @@ void Solns2Out::initFromOzn(const std::string& filename) {
   {
     _env = new Env();
     std::stringstream errstream;
-    if ((_outputModel = parse(*_env, filenames, std::vector<std::string>(), "", "", _includePaths,
-                              {}, false, false, false, false, errstream)) != nullptr) {
+    _outputModel = parse(*_env, filenames, std::vector<std::string>(), "", "", _includePaths, {},
+                         false, false, false, false, errstream);
+    if (_outputModel != nullptr) {
       std::vector<TypeError> typeErrors;
       _env->model(_outputModel);
       MZN_ASSERT_HARD_MSG(_env, "solns2out: could not allocate Env");
@@ -207,20 +206,21 @@ void Solns2Out::parseAssignments(string& solution) {
   for (unsigned int i = 0; i < sm->size(); i++) {
     if (auto* ai = (*sm)[i]->dynamicCast<AssignI>()) {
       auto& de = findOutputVar(ai->id());
-      if (!ai->e()->isa<BoolLit>() && !ai->e()->isa<IntLit>() && !ai->e()->isa<FloatLit>()) {
+      if (!Expression::isa<BoolLit>(ai->e()) && !Expression::isa<IntLit>(ai->e()) &&
+          !Expression::isa<FloatLit>(ai->e())) {
         Type de_t = de.first->type();
         de_t.cv(false);
-        ai->e()->type(de_t);
+        Expression::type(ai->e(), de_t);
       }
       ai->decl(de.first);
       typecheck(*_env, getModel(), ai);
-      if (Call* c = ai->e()->dynamicCast<Call>()) {
+      if (Call* c = Expression::dynamicCast<Call>(ai->e())) {
         // This is an arrayXd call, make sure we get the right builtin
-        assert(c->arg(c->argCount() - 1)->isa<ArrayLit>());
+        assert(Expression::isa<ArrayLit>(c->arg(c->argCount() - 1)));
         for (unsigned int i = 0; i < c->argCount(); i++) {
-          c->arg(i)->type(Type::parsetint());
+          Expression::type(c->arg(i), Type::parsetint());
         }
-        c->arg(c->argCount() - 1)->type(de.first->type());
+        Expression::type(c->arg(c->argCount() - 1), de.first->type());
         c->decl(getModel()->matchFn(_env->envi(), c, false));
       }
       de.first->e(ai->e());
@@ -236,8 +236,7 @@ void Solns2Out::declNewOutput() {
 
 void Solns2Out::printSolution(std::istream& sol, std::ostream& os, bool outputTime) {
   if (opt.flagEncapsulateJSON) {
-    std::string t = opt.flagIsChecker ? "checker" : "solution";
-    os << "{\"type\": \"" << t << "\", ";
+    os << "{\"type\": \"solution\", ";
     std::string line;
     while (std::getline(sol, line)) {
       // Remove line breaks from JSON object
@@ -305,7 +304,19 @@ bool Solns2Out::evalOutput() {
       std::string line;
       if (std::getline(checkerStream, line)) {
         if (opt.flagEncapsulateJSON) {
-          _os << line << "\n";
+          std::string last = line;
+          _os << "{\"type\": \"checker\", \"messages\": [" << line;
+          while (std::getline(checkerStream, line)) {
+            last = line;
+            _os << ", " << line;
+          }
+          _os << "]";
+          if (last.size() > 19 && last.substr(0, 19) == "{\"type\": \"solution\"") {
+            // For backwards compatibility, output solution message as part of this message
+            _os << last.substr(19) << "\n";
+          } else {
+            _os << "}\n";
+          }
         } else {
           _os << "% Solution checker report:\n";
           _os << "% " << line << "\n";
@@ -347,23 +358,25 @@ void Solns2Out::checkSolution(std::ostream& oss) {
 
   std::ostringstream checker;
   checker << _checkerModel;
+  Printer p(checker, 0, false, &getEnv()->envi());
   {
     GCLock lock;
     for (auto& i : *getModel()) {
       if (auto* vdi = i->dynamicCast<VarDeclI>()) {
-        if (vdi->e()->ann().contains(Constants::constants().ann.mzn_check_var)) {
+        if (Expression::ann(vdi->e()).contains(Constants::constants().ann.mzn_check_var)) {
           checker << vdi->e()->id()->str() << " = ";
           Expression* e = eval_par(getEnv()->envi(), vdi->e()->e());
-          auto* al = e->dynamicCast<ArrayLit>();
+          auto* al = Expression::dynamicCast<ArrayLit>(e);
           std::vector<Id*> enumids;
-          if (Call* cev = vdi->e()->ann().getCall(Constants::constants().ann.mzn_check_enum_var)) {
-            auto* enumIdsAl = cev->arg(0)->cast<ArrayLit>();
+          if (Call* cev = Expression::ann(vdi->e()).getCall(
+                  Constants::constants().ann.mzn_check_enum_var)) {
+            auto* enumIdsAl = eval_array_lit(getEnv()->envi(), cev->arg(0));
             for (int j = 0; j < enumIdsAl->size(); j++) {
-              enumids.push_back((*enumIdsAl)[j]->dynamicCast<Id>());
+              enumids.push_back(Expression::dynamicCast<Id>((*enumIdsAl)[j]));
             }
           }
-
-          if (al != nullptr) {
+          bool isArray = al != nullptr && al->type().dim() > 0;
+          if (isArray) {
             checker << "array" << al->dims() << "d(";
             for (int i = 0; i < al->dims(); i++) {
               if (!enumids.empty() && enumids[i] != nullptr) {
@@ -379,9 +392,9 @@ void Solns2Out::checkSolution(std::ostream& oss) {
           if (!enumids.empty() && enumids.back() != nullptr) {
             checker << "to_enum(" << *enumids.back() << "," << *e << ")";
           } else {
-            checker << *e;
+            p.print(e);
           }
-          if (al != nullptr) {
+          if (isArray) {
             checker << ")";
           }
           checker << ";\n";
@@ -392,6 +405,7 @@ void Solns2Out::checkSolution(std::ostream& oss) {
 
   MznSolver slv(oss, _log, _starttime);
   slv.s2out.opt.solutionSeparator = "";
+  slv.s2out.opt.searchCompleteMsg = "";
   try {
     slv.run(opt.checkerArgs, checker.str(), "minizinc", "checker.mzc");
   } catch (const Exception& e) {
@@ -490,7 +504,7 @@ bool Solns2Out::evalStatusMsg(SolverInstance::Status status) {
       }
       return true;
     case SolverInstance::OPT: {
-      if (opt.flagEncapsulateJSON) {
+      if (opt.flagEncapsulateJSON && !opt.searchCompleteMsg.empty()) {
         bool sat = getEnv()->flat()->solveItem()->st() == SolveI::ST_SAT;
         label = sat ? "ALL_SOLUTIONS" : "OPTIMAL_SOLUTION";
       } else {
@@ -521,7 +535,9 @@ bool Solns2Out::evalStatusMsg(SolverInstance::Status status) {
   }
   if (opt.flagEncapsulateJSON) {
     if (label.empty()) {
-      getOutput() << "{\"type\": \"time\", \"time\": " << _starttime.ms().count() << "}\n";
+      if (opt.flagOutputTime) {
+        getOutput() << "{\"type\": \"time\", \"time\": " << _starttime.ms().count() << "}\n";
+      }
     } else {
       getOutput() << "{\"type\": \"status\", \"status\": \"" << label << "\"";
       if (opt.flagOutputTime) {
@@ -663,7 +679,7 @@ bool Solns2Out::feedRawDataChunk(const char* data) {
               _stats.clear();
             } else {
               message << "{\"type\": \"comment\", \"comment\": \"" << Printer::escapeStringLit(line)
-                      << "\"}\n";
+                      << "\\n\"}\n";
             }
           } else {
             message << line << '\n';
